import numpy as np
import quaternion
import shapely
from shapely.geometry import Polygon
import trimesh
import pcg_gazebo
import copy

def vec_to_trans(vec):
    a = np.identity(len(vec)+1)
    a[:len(vec),len(vec)] = vec
    return a

def add_dimension(points, value=0):
    return np.concatenate([points[np.newaxis,:,i].T if i<len(points[0]) else np.full([len(points),1], value) for i in range(len(points[0])+1)], 1)
    
def get_square_horizon(base_pos, radius, z_angle=0):
    d1 = [1,1,-1,-1]
    d2 = [1,-1,-1,1]
    poss = np.array([np.quaternion(1,d1[i]*radius,d2[i]*radius,0) for i in range(4)])
    q = quaternion.from_euler_angles(0,0,z_angle)
    pq = q*poss*q.conj()
    
    return np.array([[p.x, p.y, p.z] for p in pq])+base_pos

def get_extended_face(face_vertices, length):
    face_bottom = np.copy(face_vertices)
    face_bottom[:,2] += length
    return np.concatenate([face_vertices, face_bottom])
    
def distance_filtered_poly(polygon, distance=0.01):
    c = np.round(polygon.exterior.coords, 3)
    ban = np.zeros([len(c)], dtype=int)
    for i in range(len(c)):
        if not ban[i]:
            n = np.linalg.norm(c-c[i], axis=1) < distance
            n[i] = False
            ban = np.logical_or(ban, n) 
    
    return Polygon(c[np.logical_not(ban)])

def get_initial_vec_from_model_collision(model):
    col = model.get_link_by_name(model.name).collisions[0]
    return col.pose.position

def get_corrected_poly_with_model(target_poly, model):
    col_vec = get_initial_vec_from_model_collision(model)
    corrected = shapely.affinity.affine_transform(target_poly, matrix=[1,0,0,1,col_vec[0],col_vec[1]])
    return corrected

def get_mesh_from_model(model):
    col_position = get_initial_vec_from_model_collision(model)
    col_mesh_before_trans = col.geometry._geometry_entity.get_meshes()[0]
    return col_mesh_before_trans, col_position

def get_corrected_mesh_from_model(model):    
    mesh, t = get_mesh_from_model(model)
    c_mesh = copy.deepcopy(mesh)
    c_mesh.apply_transform(vec_to_trans(t)) #-
    return c_mesh

def get_poly_from_model(model, mesh_slice_height): ## height; ROOM_WALL_HEIGHT/2
    c_mesh = get_corrected_mesh_from_model(model)
    poly = trimesh.path.polygons.projected(c_mesh, (0,0,mesh_slice_height))
    return poly

def get_interior_poly_from_extrude_mesh(mesh, mesh_slice_height):
    extrude_poly = trimesh.path.polygons.projected(mesh, (0,0,mesh_slice_height))
    return distance_filtered_poly(Polygon(extrude_poly.interiors[0]))

def get_interior_poly_from_extrude_model(model, mesh_slice_height):
    extrude_poly = get_poly_from_model(model, mesh_slice_height)
    extrude_interior_poly = distance_filtered_poly(Polygon(extrude_poly.interiors[0])) #-
    return extrude_interior_poly


import copy
from pcg_gazebo.simulation.properties.pose import Pose
from pcg_gazebo.generators.creators import create_models_from_config
from multiprocessing import Process

class ModelManager(): ## orientation: [roll, pitch, yaw] (degrees) or [qx, qy, qz, qw] (quaternion)
    
    namespace = 0
    
    def __init__(self, gazebo_proxy):
        self.gazebo_proxy = gazebo_proxy
        self.namespace = ModelManager.namespace
        ModelManager.namespace += 1
        self.configspaces = dict()
        self.modelspaces = dict()
    
    def is_set_modelspace(self, tag):
        return tag in self.modelspaces.keys()
        
    def set_modelspace_from_config(self, tag, config, max_model_count, disable_collision=False):
        configs = [copy.deepcopy(config)[0] for _ in range(max_model_count)]
        for i in range(max_model_count):
            configs[i]['args']['name'] = "mm-{}-{}_{}".format(self.namespace, tag, i)
        
        self.configspaces[tag] = configs
        models = create_models_from_config(configs)
        
        if disable_collision:
            for m in models:
                m.get_link_by_name('link').disable_collision()
        
        self.modelspaces[tag] = models
        
    def set_modelspace_from_models(self, tag, models):
        for i in range(len(models)):
            models[i].name = "mm-{}-{}_{}".format(self.namespace, tag, i)
            self.modelspaces[tag][i] = models[i]
        
    def get_moved_models(self, tag, positions, orientations):
        copy_models = copy.deepcopy(self.modelspaces[tag][:len(positions)])
        
        poses = [Pose(pos=p, rot=o) for p,o in zip(positions, orientations)]
        for i in range(len(positions)):
            copy_models[i].pose += poses[i]
        
        return copy_models
    
    def get_base_models(self, tag, count):
        return self.modelspaces[tag][0:count]
        
    def get_base_model_poses(self, tag, count):
        return [self.modelspaces[tag][i].pose for i in range(count)]
            
    def apply_model(self, tag, positions, orientations):
        self._delete_other_models()
        self._delete_models(tag, len(positions), len(self.modelspaces[tag]))
        
        spawn_f = [None] * len(positions)
        spawn_kwargs = [None] * len(positions)
        
        for i in range(len(spawn_f)):
            spawn_f[i] = self.modelspaces[tag][i].spawn
            spawn_kwargs[i] = dict(
                gazebo_proxy=self.gazebo_proxy, 
                robot_namespace=self.modelspaces[tag][i].name,
                pos=list(positions[i]),
                rot=list(orientations[i])
            )
        self._multiprocessing(spawn_f, spawn_kwargs)

    def _is_mine_model(self, model_name):
        namespace = model_name.split("-")[1]
        return ModelManager._is_manager_model(model_name) and namespace == str(self.namespace)
    
    @classmethod
    def _is_manager_model(cls, model_name):
        name = model_name.split("-")
        return name[0] == "mm" and str.isdigit(name[1])

    def _delete_models(self, tag, start, stop):     
        del_kwargs = [dict(model_name="mm-{}-{}_{}".format(self.namespace, tag, i)) for i in range(start, stop)]
        self._multiprocessing(self.gazebo_proxy.delete_model, del_kwargs)

    def _delete_other_models(self):
        manager_models = filter(ModelManager._is_manager_model, self.gazebo_proxy.get_model_names())
        other_models = filter(lambda m: not self._is_mine_model(m), manager_models)
        
        del_kwargs = [dict(model_name=m) for m in other_models]
        self._multiprocessing(self.gazebo_proxy.delete_model, del_kwargs)
            
    def _multiprocessing(self, targets, kwargs):
        if callable(targets):
            f = [targets] * len(kwargs)
        else:
            f = targets
            
        ps = [Process(target=f[i], kwargs=kwargs[i]) for i in range(len(kwargs))]
        
        for p in ps:
            p.start()
        for p in ps:
            p.join()
        while True:
            if all([p.exitcode == 0 for p in ps]):
                break


import os
import sys
import setuptools
import codecs

version_file = os.path.join(os.path.dirname(__file__), 'roomor/version.py')

with open(version_file, 'r') as f:
    __version__ = eval(f.read().strip().split('=')[-1])
    
with codecs.open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()
    
requirements = [
    'numpy-quaternion',
    'trimesh[easy]',
    'shapely',
    'scikit-learn',
    'opencv-python',
    'pcg_gazebo'
]

setuptools.setup(
    name='roomor',
    version=__version__,
    description='A Python framework for generating and deploying random room for Gazebo',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Reona Sato',
    author_email='www.shinderu.www@gmail.com',
    url='https://github.com/wwwshwww/roomor',
    packages=setuptools.find_packages(),
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.6',
    ],
    keywords='gazebo ros simulation robotics autogeneration environment reinforcement-learning pcg_gazebo',
    install_requires=requirements
)